from model_evaluation.rand_arch import Instruction, gen_arch, CPUGraph
from z3_model.z3_model import resolve_model_inst_mix, Sequence
import math as m
import itertools as it
import random as rnd

def compare_arch(diu1, dup1, diu2, dup2):
    nb_inst = len(diu1)
    assert len(diu2) == nb_inst
    nb_muop = len(diu1[0])
    assert(all(len(l) == nb_muop for l in diu1 + diu2))
    assert len(dup1) == len(dup2) == nb_muop
    nb_ports = len(dup1[0])
    assert(all(len(l) == nb_ports for l in dup1 + dup2))
    ret = True
    def same_insts(u1, u2):
        return all(diu1[i][u1] == diu2[i][u2] for i in range(nb_inst))
    def get_nb_port(u, dup):
        nbp = 0
        for p in range(nb_ports):
            if dup[u][p]:
                nbp += 1
        return nbp
    def get_nb_muop(p, dup):
        nbu = 0
        for u in range(nb_muop):
            if dup[u][p]:
                nbu += 1
        return nbu
    # For every muop in graph 1, there is a muop in graph2 that is connected to the exact same instructions
    ret = ( ret and all(any(same_insts(u1, u2) for u1 in range(nb_muop)) for u2
        in range(nb_muop)) and all(any(same_insts(u1, u2) for u2 in
            range(nb_muop)) for u1 in range(nb_muop))) 

    # For every muop in graph1, there is a muop in graph 2 that is connected to the same number of ports
    ret = (ret and all([any(get_nb_port(u1, dup1) == get_nb_port(u2, dup2) for
        u1 in range(nb_muop)) for u2 in range(nb_muop)]) and
        all([any(get_nb_port(u1, dup1) == get_nb_port(u2, dup2) for u1 in
            range(nb_muop)) for u2 in range(nb_muop)]))

    # For every port in graph1, there is a port in graph 2 that is connected to the same number of muops
    ret = (ret and all([any(get_nb_muop(p1, dup1) == get_nb_muop(p2, dup2) for
        p1 in range(nb_ports)) for p2 in range(nb_ports)]) and
        all(any(get_nb_muop(p1, dup1) == get_nb_muop(p2, dup2) for p1 in
            range(nb_ports)) for p2 in range(nb_ports)))
    return ret

def compare_arch_by_throughput(diu1, dup1, diu2, dup2):
    nb_inst = len(diu1)
    assert len(diu2) == nb_inst
    nb_muop = len(diu1[0])
    assert(all(len(l) == nb_muop for l in diu1 + diu2))
    assert len(dup1) == len(dup2) == nb_muop
    nb_ports = len(dup1[0])
    assert(all(len(l) == nb_ports for l in dup1 + dup2))
    basename = "inst_" 
    loc_insts = []
    for i in range(nb_inst):
        loc_insts.append(Instruction(basename + str(i)))
    graph1 = CPUGraph(loc_insts, nb_muop, nb_ports, diu1, dup1)
    graph2 = CPUGraph(loc_insts, nb_muop, nb_ports, diu2, dup2)
    mixs = []
    for i in loc_insts:
        mixs.append({i: 1})

    for (i1, i2) in it.combinations(loc_insts, 2):
        for _ in range(4):
            v1 = rnd.randint(1, 5)
            v2 = rnd.randint(1, 5)
            mixs.append({i1: v1, i2: v2})

    for mix in mixs:
        if graph1.get_throughput(mix) != graph2.get_throughput(mix):
            print("NOT EQUAL: mix {} original graph: {} guessed graph: {}".format([(i.name, mix[i]) for i in mix],graph1.get_throughput(mix), graph2.get_throughput(mix)))
            return False
    return True




if __name__ == "__main__":
    nb_muops = 2
    nb_ports = 2

    add = Instruction("add")
    mul = Instruction("mul")
    insts = [add, mul]
    mix1 = {add: 1}
    mix2 = {mul: 1}
    mix3 = {add: 1, mul: 1}
    mix4 = {add: 2, mul: 1}
    mix5 = {add: 1, mul: 2}
    mix6 = {add: 3, mul: 1}
    mix7 = {add: 1, mul: 3}
    mix8 = {add: 3, mul: 2}
    mix9 = {add: 2, mul: 3}

    nb_rights = 0
    nb_thr_rights = 0
    for _ in range(50):
        diu, dup = gen_arch([add, mul], nb_muops, nb_ports)
        graph = CPUGraph(insts, nb_muops, nb_ports, diu, dup)


        seq1 = Sequence(mix1, graph.get_throughput(mix1))
        seq2 = Sequence(mix2, graph.get_throughput(mix2))
        seq3 = Sequence(mix3, graph.get_throughput(mix3))
        seq4 = Sequence(mix4, graph.get_throughput(mix4))
        seq5 = Sequence(mix5, graph.get_throughput(mix5))
        seq6 = Sequence(mix6, graph.get_throughput(mix6))
        seq7 = Sequence(mix7, graph.get_throughput(mix7))
        seq8 = Sequence(mix8, graph.get_throughput(mix8))
        seq9 = Sequence(mix9, graph.get_throughput(mix9))

        multi = dict()
        #multi[add] = m.ceil(graph.get_throughput(mix1))
        #multi[mul] = m.ceil(graph.get_throughput(mix2))
        multi[add] = graph.get_throughput(mix1)
        multi[mul] = graph.get_throughput(mix2)
        seqs = [seq1, seq2, seq3, seq4, seq5, seq6, seq7, seq8, seq9]

        mdiu, mdup = resolve_model_inst_mix(seqs, insts, nb_muops, nb_ports, multi)
        thr_false = []

        print("\n-------------------------------")
        if compare_arch(diu, dup, mdiu, mdup):
            nb_rights += 1
        if compare_arch_by_throughput(diu, dup, mdiu, mdup):
            print("Graph OK !!")
            thr_false.append(((diu, dup), (mdiu, mdup)))
            nb_thr_rights += 1
        else:
            print("Original graph: \n\t{} \n\t {}\n Built graph \n\t{}\n\t{}".format(diu, dup, mdiu, mdup))
            
    print(nb_rights)
    print("percentage right graph: {}%".format(nb_rights / 50 * 100))
    print(nb_thr_rights)
    print("percentage right throughput: {}%".format( float(nb_thr_rights) / 50 * 100))


